/* dbus_common.h
 *
 * Copyright (C) 2013 Canonical, Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include <dbus/dbus.h>

const char *type_to_name(int message_type);
void log_message(int log_fd, const char *prefix, DBusMessage * message);
void append_arg(DBusMessageIter * iter, int type, const char *value);
void append_array(DBusMessageIter * iter, int type, const char *value);
void append_dict(DBusMessageIter * iter, int keytype, int valtype,
		 const char *value);
int type_from_name(const char *arg);
